"""Helper for ASAM ODS HTTP API conI session"""

import logging
from typing import List

import requests
import requests.auth
from google.protobuf.json_format import MessageToJson
from google.protobuf.message import Message
from pandas import DataFrame

import odsbox.proto.ods_pb2 as _ods
from odsbox.datamatrices_to_pandas import to_pandas
from odsbox.jaquel import jaquel_to_ods
from odsbox.model_cache import ModelCache
from odsbox.transaction import Transaction


class ConI:
    """This is a helper to hold an ASAM ODS HTTP API ConI session"""

    __log: logging.Logger = logging.getLogger(__name__)
    __session: requests.Session = None
    __con_i: str = None
    mc: ModelCache = None

    def __init__(
        self,
        url: str = "http://localhost:8080/api",
        auth: requests.auth.AuthBase = ("sa", "sa"),
        context_variables: _ods.ContextVariables | dict | None = None,
        verify_certificate: bool = True,
    ):
        session = requests.Session()
        session.auth = auth
        session.headers.update(
            {
                "Content-Type": "application/x-asamods+protobuf",
                "Accept": "application/x-asamods+protobuf",
            }
        )
        session.verify = verify_certificate
        session.timeout = 600.0

        _context_variables = None
        if isinstance(context_variables, _ods.ContextVariables):
            _context_variables = context_variables
        else:
            _context_variables = _ods.ContextVariables()
            if isinstance(context_variables, dict):
                for key, value in context_variables.items():
                    _context_variables.variables[key].string_array.values.append(value)

        response = session.post(url + "/ods", data=_context_variables.SerializeToString())
        if 201 == response.status_code:
            con_i = response.headers["location"]
            self.__log.debug("ConI: %s", con_i)
            self.__session = session
            self.__con_i = con_i
        self.__check_result(response)
        # lets cache the model
        self.model_read()

    def __del__(self):
        if None is not self.__session:
            self.logout()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        self.logout()

    def con_i_url(self):
        """Get the ASAM ODS session URL used to work with this session."""
        return self.__con_i

    def logout(self):
        """Close the attached session at the ODS server."""
        if None is not self.__session:
            response = self.__session.delete(
                self.__con_i,
            )
            self.__session.close()
            self.__session = None
            self.__con_i = None
            self.__check_result(response)

    def query_data(self, query: str | dict | _ods.SelectStatement) -> DataFrame:
        """Query ods server for content"""
        if isinstance(query, _ods.SelectStatement):
            return to_pandas(self.data_read(self.data_read_jaquel(query)))
        else:
            return to_pandas(self.data_read_jaquel(query))

    def model(self) -> _ods.Model:
        """Get the cache ODS server model."""
        return self.mc.model()

    def data_read_jaquel(self, jaquel: str | dict) -> _ods.DataMatrices:
        """Query ODS server using JAQueL syntax and return result as pandas DataFrame."""
        _ods_entity, ods_query = jaquel_to_ods(self.model(), jaquel)
        return self.data_read(ods_query)

    def data_read(self, data: _ods.SelectStatement) -> _ods.DataMatrices:
        """Query ASAM ODS server."""
        response = self.ods_post_request("data-read", data)
        return_value = _ods.DataMatrices()
        return_value.ParseFromString(response.content)
        return return_value

    def data_create(self, data: _ods.DataMatrices) -> List[int]:
        """Create new Instances."""
        response = self.ods_post_request("data-create", data)
        return_value = _ods.DataMatrices()
        return_value.ParseFromString(response.content)
        return return_value.matrices[0].columns[0].longlong_array.values

    def data_update(self, data: _ods.DataMatrices) -> None:
        """Update existing instances."""
        self.ods_post_request("data-update", data)

    def data_delete(self, data: _ods.DataMatrices) -> None:
        """Delete Instances."""
        self.ods_post_request("data-delete", data)

    def data_copy(self, data: _ods.CopyRequest) -> _ods.Instance:
        """Copy an Instance and its related children."""
        response = self.ods_post_request("data-copy", data)
        return_value = _ods.Instance()
        return_value.ParseFromString(response.content)
        return return_value

    def n_m_relation_read(self, data: _ods.NtoMRelationIdentifier) -> _ods.NtoMRelatedInstances:
        """Read n-m relations for a defined instance."""
        response = self.ods_post_request("n-m-relation-read", data)
        return_value = _ods.NtoMRelatedInstances()
        return_value.ParseFromString(response.content)
        return return_value

    def n_m_relation_write(self, data: _ods.NtoMWriteRelatedInstances) -> None:
        """Update, delete or create n-m relations for given instance pairs."""
        self.ods_post_request("n-m-relation-write", data)

    def transaction(self) -> Transaction:
        """Open a transaction object to be used in a with clause"""
        return Transaction(self)

    def transaction_create(self) -> None:
        """Open a transaction for writing."""
        self.ods_post_request("transaction-create")

    def transaction_commit(self) -> None:
        """Commit transaction created before."""
        self.ods_post_request("transaction-commit")

    def transaction_abort(self) -> None:
        """Abort transaction created before."""
        self.ods_post_request("transaction-abort")

    def valuematrix_read(self, data: _ods.ValueMatrixRequestStruct) -> _ods.DataMatrices:
        """Read bulk data from a submatrix or measurement.
        Submatrix access can also be done using data-read."""
        response = self.ods_post_request("valuematrix-read", data)
        return_value = _ods.DataMatrices()
        return_value.ParseFromString(response.content)
        return return_value

    def model_read(self) -> _ods.Model:
        """Read the model from server and update cached version."""
        response = self.ods_post_request("model-read")
        model = _ods.Model()
        model.ParseFromString(response.content)
        self.mc = ModelCache(model)
        return model

    def model_update(self, model_parts: _ods.Model, update_model: bool = True) -> None:
        """Update application model content."""
        self.ods_post_request("model-update", model_parts)
        if update_model:
            # cache again if successfully changed
            self.model_read()

    def model_delete(self, model_parts: _ods.Model, update_model: bool = True) -> None:
        """Delete application model content."""
        self.ods_post_request("model-delete", model_parts)
        if update_model:
            # cache again if successfully changed
            self.model_read()

    def model_check(self) -> None:
        """Check if stored application model is consistent."""
        self.ods_post_request("model-check")

    def basemodel_read(self) -> _ods.BaseModel:
        """Read the ODS base model version used by the server."""
        response = self.ods_post_request("basemodel-read")
        base_model = _ods.BaseModel()
        base_model.ParseFromString(response.content)
        return base_model

    def asampath_create(self, data: _ods.Instance) -> _ods.AsamPath:
        """Create an persistent string representing the instance."""
        response = self.ods_post_request("asampath-create", data)
        return_value = _ods.AsamPath()
        return_value.ParseFromString(response.content)
        return return_value

    def asampath_resolve(self, data: _ods.AsamPath) -> _ods.Instance:
        """Use the persistent string to get back the instance."""
        response = self.ods_post_request("asampath-resolve", data)
        return_value = _ods.Instance()
        return_value.ParseFromString(response.content)
        return return_value

    def context_read(self, pattern_or_filter: _ods.ContextVariablesFilter | str = "*") -> _ods.ContextVariables:
        """Read the session context variables."""
        context_variables_filter = (
            pattern_or_filter
            if isinstance(pattern_or_filter, _ods.ContextVariablesFilter)
            else _ods.ContextVariablesFilter(pattern=pattern_or_filter)
        )
        response = self.ods_post_request("context-read", context_variables_filter)
        return_value = _ods.ContextVariables()
        return_value.ParseFromString(response.content)
        return return_value

    def context_update(self, data: _ods.ContextVariables):
        """Set context variables for current session."""
        self.ods_post_request("context-update", data)

    def password_update(self, data: _ods.PasswordUpdate) -> None:
        """Update the password of the defined user."""
        self.ods_post_request("password-update", data)

    def ods_post_request(self, url: str, data: Message | None = None) -> requests.Response:
        """Do ODS post call with the given relative URL."""

        response = self.__session.post(
            self.__con_i + "/" + url,
            data=data.SerializeToString() if data is not None else None,
        )
        self.__check_result(response)
        return response

    def __check_result(self, response: requests.Response):
        if response.status_code not in (200, 201):
            response.headers
            if (
                "Content-Type" in response.headers
                and "application/x-asamods+protobuf" == response.headers["Content-Type"]
            ):
                error_info = _ods.ErrorInfo()
                error_info.ParseFromString(response.content)
                raise requests.HTTPError(
                    MessageToJson(error_info),
                    response=response,
                )
            response.raise_for_status()



import logging

import odsbox.proto.ods_pb2 as ods
from odsbox.datamatrices_to_pandas import to_pandas


def test_conversion1():
    dms = ods.DataMatrices()
    dm = dms.matrices.add(aid=4711, name="MeaQuantity")
    dm.columns.add(name="Name").string_array.values[:] = ["Name"] * 13
    dm.columns.add(name="Description").string_array.values[:] = ["Description"] * 13
    dm.columns.add(name="MimeType").string_array.values[:] = ["application/x-asam.aomeasurementquantity"] * 13
    dm.columns.add(name="DataType").long_array.values[:] = [ods.DT_DOUBLE] * 13
    dm.columns.add(name="MeaResult").longlong_array.values[:] = [4611] * 13
    dm.columns.add(name="Unit").longlong_array.values[:] = [15] * 13
    dm.columns.add(name="Maximum").double_array.values[:] = [1.2] * 13
    dm.columns.add(name="Minimum").double_array.values[:] = [0.9] * 13

    pdf = to_pandas(dms)
    logging.getLogger().info(pdf)
    assert pdf.shape == (13, 8)


def test_byte_array():
    dms = ods.DataMatrices()
    dm = dms.matrices.add(aid=4711, name="MeaQuantity")
    dm.columns.add(name="byte_val").byte_array.values = bytes([1] * 13)

    pdf = to_pandas(dms)
    logging.getLogger().info(pdf)
    assert pdf.shape == (13, 1)


def test_all_types():
    dms = ods.DataMatrices()
    dm = dms.matrices.add(aid=4711, name="AllTypes")
    dm.columns.add(name="Str").string_array.values[:] = ["StrVal"] * 2
    dm.columns.add(name="I32").long_array.values[:] = [1] * 2
    dm.columns.add(name="F32").float_array.values[:] = [1.2] * 2
    dm.columns.add(name="Bool").boolean_array.values[:] = [True] * 2
    dm.columns.add(name="U08").byte_array.values = b"ab"
    dm.columns.add(name="F64").double_array.values[:] = [2.2] * 2
    dm.columns.add(name="I64").longlong_array.values[:] = [1] * 2

    pdf = to_pandas(dms)
    logging.getLogger().info(pdf)
    assert pdf.shape == (2, 7)



"""Test ModelCache functionality."""

from __future__ import annotations

import os
from pathlib import Path
import pytest

from google.protobuf.json_format import Parse

from odsbox.model_cache import ModelCache
from odsbox.proto.ods_pb2 import Model


def __get_model(model_file_name):
    model_file = os.path.join(os.path.abspath(os.path.dirname(__file__)), "test_data", model_file_name)
    model = Model()
    Parse(Path(model_file).read_text(encoding="utf-8"), model)
    return model


def test_model_cache():
    model = __get_model("mdm_nvh_model.json")

    mc = ModelCache(model)
    local_column_entity = mc.entity_by_base_name("AoLocalColumn")
    assert 82 == local_column_entity.aid
    assert "LocalColumn" == local_column_entity.name
    assert "AoLocalColumn" == local_column_entity.base_name
    assert "LocalColumn" == mc.entity_by_base_name("aolocalcolumn").name
    assert 82 == mc.aid_by_entity_name("LocalColumn")

    id_attribute = mc.attribute_by_base_name(local_column_entity, "id")
    assert "id" == id_attribute.base_name
    assert "Id" == id_attribute.name
    assert "id" == mc.attribute_by_base_name(local_column_entity, "ID").base_name
    assert "id" == mc.attribute_by_base_name("LocalColumn", "id").base_name
    assert "id" == mc.attribute_by_base_name("LocalColumn", "ID").base_name

    submatrix_relation = mc.relation_by_base_name(local_column_entity, "submatrix")
    assert "submatrix" == submatrix_relation.base_name
    assert "SubMatrix" == submatrix_relation.name
    assert "submatrix" == mc.relation_by_base_name(local_column_entity, "SUBMATRIX").base_name
    assert "submatrix" == mc.relation_by_base_name("LocalColumn", "submatrix").base_name
    assert "submatrix" == mc.relation_by_base_name("LocalColumn", "SUBMATRIX").base_name


def test_enumeration_access():
    mc = ModelCache(__get_model("mdm_nvh_model.json"))
    assert mc.enumeration("datatype_enum") is not None
    assert mc.enumeration("DataType_Enum") is not None
    with pytest.raises(ValueError):
        mc.enumeration("DoesNotExist")


def test_enumeration_values():
    mc = ModelCache(__get_model("mdm_nvh_model.json"))
    datatpype_enum = mc.enumeration("datatype_enum")
    assert datatpype_enum is not None

    assert 3 == mc.enumeration_key_to_value(datatpype_enum, "DT_FLOAT")
    assert 3 == mc.enumeration_key_to_value(datatpype_enum, "dt_float")
    assert "DT_FLOAT" == mc.enumeration_value_to_key(datatpype_enum, 3)

    assert 3 == mc.enumeration_key_to_value("datatype_enum", "DT_FLOAT")
    assert 3 == mc.enumeration_key_to_value("datatype_enum", "dt_float")
    assert "DT_FLOAT" == mc.enumeration_value_to_key("datatype_enum", 3)

    assert 3 == mc.enumeration_key_to_value("DATATYPE_ENUM", "DT_FLOAT")
    assert 3 == mc.enumeration_key_to_value("DATATYPE_ENUM", "dt_float")
    assert "DT_FLOAT" == mc.enumeration_value_to_key("DATATYPE_ENUM", 3)

    with pytest.raises(ValueError):
        assert 3 == mc.enumeration_key_to_value("DOESNOTEXIST", "DT_FLOAT")
    with pytest.raises(ValueError):
        assert 3 == mc.enumeration_key_to_value("DATATYPE_ENUM", "DOESNOTEXIST")
    with pytest.raises(ValueError):
        assert "DT_FLOAT" == mc.enumeration_value_to_key("DOESNOTEXIST", 3)
    with pytest.raises(ValueError):
        assert "DT_FLOAT" == mc.enumeration_value_to_key("DATATYPE_ENUM", 501)



"""helps working with the ASAM ODS model"""

import logging

import odsbox.proto.ods_pb2 as _ods


class ModelCache:
    """The model object returned from ods server needs some utilities to work with it."""

    __model: _ods.Model = None
    __log: logging.Logger = logging.getLogger(__name__)

    def __init__(self, model: _ods.Model):
        self.__model = model

    def model(self) -> _ods.Model:
        """Get the attached ASAM ODS model"""
        return self.__model

    def aid_by_entity_name(self, entity_name: str) -> int:
        """
        Determine the application element id of an entity by its name.

        :entity_name: Case sensitive name to look for
        :raises ValueError: If the entity does not exist.
        """
        if entity_name in self.__model.entities:
            return self.__model.entities[entity_name].aid
        raise ValueError(f"Entity {entity_name} does not exist in model")

    def entity_by_base_name(self, entity_base_name: str) -> _ods.Model.Entity:
        """
        Get the entity by its base name.

        :entity_base_name: case insensitive name of the base model element.
        :raises ValueError: If the entity does not exist.
        """
        entities = self.__model.entities
        for key in entities:
            entity = entities[key]
            if entity_base_name.casefold() == entity.base_name.casefold():
                return entity
        raise ValueError(f"No entity derived from {entity_base_name}")

    def relation_by_base_name(
        self, entity_or_name: str | _ods.Model.Entity, relation_base_name: str
    ) -> _ods.Model.Relation:
        """
        Get the relation by base name.

        :entity_or_name: entity object or case sensitive application name to lookup
        :relation_base_name: case insensitive name of the base model element.
        :raises ValueError: If the relation does not exist.
        """
        entity = self.__entity(entity_or_name)
        relations = entity.relations
        for key in relations:
            relation = relations[key]
            if relation_base_name.casefold() == relation.base_name.casefold():
                return relation
        raise ValueError(f"Entity {entity.name} does not have relation derived from {relation_base_name}")

    def relation_name_by_base_name(self, entity_or_name: str | _ods.Model.Entity, relation_base_name: str) -> str:
        """
        Get the relation application name by base name.

        :entity_or_name: entity object or case sensitive application name to lookup
        :relation_base_name: case insensitive name of the base model element.
        :raises ValueError: If the relation does not exist.
        """
        return self.relation_by_base_name(entity_or_name, relation_base_name).name

    def attribute_by_base_name(
        self, entity_or_name: str | _ods.Model.Entity, attribute_base_name: str
    ) -> _ods.Model.Attribute:
        """
        Get the attribute by base name.

        :entity_or_name: entity object or case sensitive application name to lookup
        :attribute_base_name: case insensitive name of the base model element.
        :raises ValueError: If the attribute does not exist.
        """
        entity = self.__entity(entity_or_name)
        attributes = entity.attributes
        for key in attributes:
            attribute = attributes[key]
            if attribute_base_name.casefold() == attribute.base_name.casefold():
                return attribute
        raise ValueError(f"Entity {entity.name} does not have attribute derived from {attribute_base_name}")

    def attribute_name_by_base_name(self, entity_or_name: str | _ods.Model.Entity, attribute_base_name: str) -> str:
        """
        Get the attribute by base name.

        :entity_or_name: entity object or case sensitive application name to lookup
        :attribute_base_name: case insensitive name of the base model element.
        :raises ValueError: If the attribute does not exist.
        """
        return self.attribute_by_base_name(entity_or_name, attribute_base_name).name

    def enumeration(self, enumeration_name: str) -> _ods.Model.Enumeration:
        """
        Get enumeration by its name.

        :enumeration_name: case insensitive name of the application model enumeration.
        :raises ValueError: If the enumeration does not exist.
        """
        if enumeration_name in self.__model.enumerations:
            return self.__model.enumerations[enumeration_name]

        for key in self.__model.enumerations:
            if key.casefold() == enumeration_name.casefold():
                return self.__model.enumerations[key]
        raise ValueError(f"Enumeration {enumeration_name} does not exist in datamodel")

    def enumeration_value_to_key(self, enumeration_or_name: str | _ods.Model.Enumeration, lookup_value: int):
        """
        Convert an enumeration value into its string representation.

        :enumeration_or_name: ods enumeration or its case insensitive name
        :lookup_value: integer value to check
        :raises ValueError: If the enumeration does not exist or does not contain value.
        """
        enumeration = (
            enumeration_or_name
            if isinstance(enumeration_or_name, _ods.Model.Enumeration)
            else self.enumeration(enumeration_or_name)
        )
        for key, value in enumeration.items.items():
            if value == lookup_value:
                return key
        raise ValueError(f"Enumeration {enumeration.name} does not contain the int value {lookup_value}")

    def enumeration_key_to_value(self, enumeration_or_name: str | _ods.Model.Enumeration, lookup_key: str):
        """
        Convert an enumeration integer value into its string representation.

        :enumeration_or_name: ods enumeration or its case insensitive name
        :lookup_key: case insensitive string key value to check
        :raises ValueError: If the enumeration does not exist or does not contain the key.
        """
        enumeration = (
            enumeration_or_name
            if isinstance(enumeration_or_name, _ods.Model.Enumeration)
            else self.enumeration(enumeration_or_name)
        )

        if lookup_key in enumeration.items:
            return enumeration.items[lookup_key]

        for key, value in enumeration.items.items():
            if key.casefold() == lookup_key.casefold():
                return value
        raise ValueError(f"Enumeration {enumeration.name} does not contain the key {lookup_key}")

    def __entity(self, entity_or_name: str | _ods.Model.Entity) -> _ods.Model.Entity:
        rv = entity_or_name if isinstance(entity_or_name, _ods.Model.Entity) else self.__model.entities[entity_or_name]
        if rv is None:
            raise ValueError(f"No entity named {entity_or_name} found")
        return rv



# pylint: disable=C0114, C0115, C0116, E1101
import os
import json
from pathlib import Path
import logging

from google.protobuf.json_format import MessageToJson, Parse

from odsbox.jaquel import jaquel_to_ods
import odsbox.proto.ods_pb2 as ods


def __get_model(model_file_name):
    model_file = os.path.join(os.path.abspath(os.path.dirname(__file__)), "test_data", model_file_name)
    model = ods.Model()
    Parse(Path(model_file).read_text(encoding="utf-8"), model)
    return model


def test_conversion1():
    model = __get_model("mdm_nvh_model.json")

    entity, ss = jaquel_to_ods(
        model,
        {
            "AoMeasurement": {
                "$or": [
                    {"measurement_quantities.maximum": {"$gte": 1, "$lt": 2}},
                    {"measurement_quantities.maximum": {"$gte": 3, "$lt": 4}},
                    {"measurement_quantities.maximum": {"$gte": 6, "$lt": 7}},
                ]
            },
            "$options": {"$rowlimit": 1000, "$rowskip": 500, "$seqlimit": 1000, "$seqskip": 500},
            "$attributes": {"name": 1, "id": 1, "test": {"name": 1, "id": 1}},
            "$orderby": {"name": 1},
        },
    )
    logging.getLogger().info(MessageToJson(ss))
    assert entity is not None
    assert entity.name == "MeaResult"
    assert ss is not None


def test_conversion2():
    model = __get_model("mdm_nvh_model.json")

    entity, ss = jaquel_to_ods(
        model,
        """{
    "AoMeasurement": {
        "measurement_begin": {
            "$between": [
                "2012-04-22T00:00:00.010000Z",
                "2012-04-23T00:00:00.000000Z"
            ]
        }
    }
}""",
    )
    logging.getLogger().info(MessageToJson(ss))
    assert entity is not None
    assert entity.name == "MeaResult"
    assert ss is not None


def __read_json_file(file_path):
    with open(file_path, encoding="utf-8") as fh:
        return json.load(fh)


def test_predefined():
    predefined_path = os.path.join(os.path.abspath(os.path.dirname(__file__)), "test_data", "jaquel")

    model = __get_model("mdm_nvh_model.json")

    for path in Path(predefined_path).rglob("*.json"):
        logging.getLogger().info(path.stem)

        jaquel_dict = __read_json_file(path)

        _entity, select_statement = jaquel_to_ods(model, jaquel_dict)

        select_statement_ref = ods.SelectStatement()
        Parse(
            Path(str(path) + ".proto").read_text(encoding="utf-8"),
            select_statement_ref,
        )
        assert select_statement_ref == select_statement
